#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This application is released under the GNU General Public License 
# v3 (or, at your option, any later version). You can find the full 
# text of the license under http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
# Thank you for using free software!

# PhantScreenlet (c) Andi Hasenkopf 2010 <andi@hasenkopf2000.net>
# Default icon by DBGthekafu is licensed under the GPL. 
# http://gnome-look.org/usermanager/search.php?username=DBGthekafu

import screenlets
from screenlets.options import StringOption , BoolOption , IntOption , FileOption , DirectoryOption , ListOption , AccountOption , TimeOption , FontOption, ColorOption , ImageOption
from screenlets.options import create_option_from_node
from screenlets import DefaultMenuItem
import pango
import gobject
import gtk
import datetime
from xml.etree import ElementTree
import gdata.calendar.service
import gdata.service
import gdata.calendar
import atom
import re
import subprocess
import urllib
import os.path
try:
	import pynotify
except:
	pynotify_installed = False
else:
	pynotify_installed = True

class PhantCalendarScreenlet (screenlets.Screenlet):
	"""Screenlet displaying upcoming events from your Google Calendar"""
	
	# default meta-info for Screenlets (should be removed and put into metainfo)
	__name__	= 'PhantCalendarScreenlet'
	__version__	= '1.0'
	__author__	= 'Andi Hasenkopf'
	__desc__	= __doc__	# set description to docstring of class
	
	# editable options (options that are editable through the UI)
	icon = 'logo'
	backgr = 'back'
	empty_acc = ('','')
	actual_acc = ('','')
	displayed_weeks = 2
	font = "Sans\ 10"
	font_color = (1,1,1, 0.8)
	board_color_0 = (.1,.1,.1, 0.8)
	board_color_1 = (.05,.05,.05, 0.8)
	shadow_color = (0,0,0, 0.6)
	highlight_color_0 = (1,0.7,0.7,0.8)
	highlight_color_1 = (0.7,1,0.7,0.8)
	appointment = []
	file = 'calendar.csv'
	on_line = False
	logged_in = False
	offset = 0
	notify_available = pynotify_installed
	notify_activated = False
	no_of_fails = 0

	# calendar settings
	feedPrefix = 'http://www.google.com/calendar/feeds/'

	# constructor
	def __init__ (self, **keyword_args):
		#call super (width/height MUST match the size of graphics in the theme)
		screenlets.Screenlet.__init__(self, width=int(218*self.scale), 
			height=int(318*self.scale), uses_theme=True, **keyword_args)
		# set theme
		self.theme_name = "default"

		self.add_options_group('Feed', 'Configure the Google Feed')
		self.add_options_group('Board', 'Configure the colors, etc.')

		# add editable option to the group
		self.Account = AccountOption('Feed','actual_acc',
			self.empty_acc, 'Username and Password', 'Using keyring encryption')
		self.add_option(self.Account)

		self.Displayed_Weeks = IntOption('Feed','displayed_weeks',
			self.displayed_weeks,'# weeks', 'Number of weeks to look into future',0,10)
		self.add_option(self.Displayed_Weeks)

		self.add_option(FontOption('Board','font', 
			self.font, 'Font', 
			'This font will be used.'))

		self.add_option(ColorOption('Board','font_color', 
			self.font_color, 'Font color', 
			'This is the font\'s color'))

		self.add_option(ColorOption('Board','board_color_0', 
			self.board_color_0, 'Board color #1', 
			'This is the board\'s color #1'))

		self.add_option(ColorOption('Board','board_color_1', 
			self.board_color_1, 'Board color #2', 
			'This is the board\'s color #2'))

		self.add_option(ColorOption('Board','shadow_color', 
			self.shadow_color, 'Shadow color', 
			'This is the shadow color'))

		self.add_option(ColorOption('Board','highlight_color_0', 
			self.highlight_color_0, 'Highlight color 1', 
			'This color is used for highlighting events'))

		self.add_option(ColorOption('Board','highlight_color_1', 
			self.highlight_color_1, 'Highlight color 2', 
			'This color is used for highlighting events'))

		if self.notify_available:
			self.add_option(BoolOption('Feed','notify_activated', 
				self.notify_activated, 'Notifications?', 
				'Activate Desktop Notifications'))

	def verify_connection ( self ):
		ping = subprocess.Popen(['ping','-c3','www.google.com'],stdout=subprocess.PIPE)
		reply = ping.communicate()[0]
		match = re.search('([0-9]+)[a-z ,]+transmitted[a-z ,]+([0-9]+)[a-z ,]+received',reply)
		if match:
			if int(match.group(1)) == 3 and int(match.group(2)) > 0:
				self.on_line = True
		else:
			self.on_line = False
			self.logged_in = False
		return self.on_line

	def calendar_login ( self ):	
		self.calendar_service = gdata.calendar.service.CalendarService()
		self.calendar_service.email = self.actual_acc[0]
		self.calendar_service.password = self.actual_acc[1]
		self.calendar_service.source = 'Google-Calendar_Python_Sample-1.0'
		try:
			self.calendar_service.ProgrammaticLogin()
		except gdata.service.CaptchaRequired:
			captcha_token = self.calendar_service._GetCaptchaToken()
			captcha_url = self.calendar_service._GetCaptchaURL()
			captcha_img = gtk.Image()
			
			# Load the actual captacha image to the GTK Image object
			raw_img = urllib.urlopen(captcha_url)
			loader = gtk.gdk.PixbufLoader()
			loader.write(raw_img.read())
			loader.close()
			captcha_img.set_from_pixbuf(loader.get_pixbuf())
			
			response = gtk.Entry()
			
			diag = gtk.Dialog(title="Captcha Required")
			diag.vbox.add(gtk.Label("The Google Calendar API\nis asking for a Captcha"))
			diag.vbox.add(gtk.Label("Username: %s" % self.actual_acc[0]))
			diag.vbox.add(captcha_img)
			diag.vbox.add(response)
			diag.add_button(gtk.STOCK_OK,1)
			diag.show_all()
			diag.run()
			
			captcha_text = response.get_text()
			
			try:
				self.calendar_service.ProgrammaticLogin(captcha_token,captcha_text)
			except:
				if self.notify_activated:
					n = pynotify.Notification(
						"PhantCalendar Oops!",
						"Please try again to login\nor solve the Captcha", 
						gtk.STOCK_DIALOG_ERROR)
					n.show()
				else:
					print "\033[43mFailed to LOGIN!\033[0m"
				self.logged_in=False
			else:
				if self.notify_activated:
					n = pynotify.Notification(
						"PhantCalendar",
						"Login successful", 
						gtk.STOCK_DIALOG_INFO)
					n.show()
				else:
					print "\033[42mLOGIN successfull!\033[0m"
				self.logged_in=True
			
		except gdata.service.BadAuthentication:
			if self.notify_activated:
				n = pynotify.Notification(
					"PhantCalendar Error",
					"Username and/or password wrong", 
					gtk.STOCK_DIALOG_ERROR)
				n.show()
			else:
				print "\033[43mFailed to LOGIN!\033[0m"
			self.logged_in = False		
		else:
			if self.notify_activated:
				n = pynotify.Notification(
					"PhantCalendar",
					"Login successful", 
					gtk.STOCK_DIALOG_INFO)
				n.show()
			else:
				print "\033[42mLOGIN successfull!\033[0m"
			self.logged_in = True
			self.allCals = self.calendar_service.GetAllCalendarsFeed()

		return self.logged_in

	def calendar_query(self, start_date, end_date):
		if self.logged_in:
			apps = {}
			for cal in self.allCals.entry:
				cal.altLink = cal.GetAlternateLink().href
				match = re.match('^' + self.feedPrefix + '(.*?)/(.*?)/(.*)$',
	                             cal.altLink)
				cal.username    = urllib.unquote(match.group(1))
				cal.visibility  = urllib.unquote(match.group(2))
				cal.projection  = urllib.unquote(match.group(3))
				query = gdata.calendar.service.CalendarEventQuery(cal.username,cal.visibility,cal.projection)
				query.start_min = start_date
				query.start_max = end_date
				feed = self.calendar_service.CalendarQuery(query)
				for i, an_event in enumerate(feed.entry):
					if an_event.when:
						try:
							date, time = an_event.when[0].start_time.split('T')
							the_datetime = datetime.datetime(int(date[0:4]),int(date[5:7]),int(date[8:10]),int(time[0:2]),int(time[3:5]))
							time = time[:5]
						except:
							date, time = an_event.when[0].start_time, None
							the_datetime = datetime.datetime(int(date[0:4]),int(date[5:7]),int(date[8:10]))
						finally:
							apps[the_datetime.strftime('%Y-%m-%d\t%H:%M')] = [date, time, an_event.title.text,cal.color.value]

			k = apps.keys()
			k.sort()
			apps = [apps[key] for key in k]
			f = open(os.path.join(self.get_screenlet_dir(),self.file),'w')
			for app in apps:
				if app[1] != None:
					f.write('%s\t%s\t%s\n' % (app[0],app[1],app[2]))
				else:
					f.write('%s\tNone\t%s\n' % (app[0],app[2]))
			f.close()
		else:
			try:
				f = open(os.path.join(self.get_screenlet_dir(),self.file),'r')
			except:
				return [[0,0,0],]
			else:
				apps = f.readlines()
				f.close()
				for i in range(len(apps)):
					apps[i] = apps[i].split('\t')
		return apps

	def notify (self):
		if pynotify.init("PhantCalendarScreenlet"):
			N = datetime.datetime.now()
			apps = []
			for app in self.appointment:
				if app[1]:
					tmp = app[0].split('-')
					tmp += app[1].split(':')
					try:
						D = datetime.datetime(int(tmp[0]),int(tmp[1]),int(tmp[2]),int(tmp[3]),int(tmp[4]))
						if D-N < datetime.timedelta(hours=1) and D-N > datetime.timedelta(microseconds=1):
							apps.append(app[2])
					except:
						# This is the case, if no data sets are present at all.
						# Therefore this try except block was added to
						# avoid ugly crashes due tu ValueErrors.
						pass
			if len(apps):			
				n = pynotify.Notification(
					"PhantCalendar Upcoming Events",
					"%s" % '\n'.join(apps), 
					gtk.STOCK_DIALOG_WARNING)
				n.show()

	def update (self):

		if self.verify_connection():
			if not self.logged_in:
				self.calendar_login()
		
		self.appointment = self.calendar_query(start_date=str(datetime.date.today()), end_date=str(datetime.date.today()+datetime.timedelta(weeks=self.displayed_weeks)))

		self.redraw_canvas()
		if self.notify_activated:
			self.notify()
		return 1
	
	# ONLY FOR TESTING!!!!!!!!!
	def init_options_from_metadata (self):
		"""Try to load metadata-file with options. The file has to be named
		like the Screenlet, with the extension ".xml" and needs to be placed
		in the Screenlet's personal directory. 
		NOTE: This function always uses the metadata-file relative to the 
			  Screenlet's location, not the ones in SCREENLETS_PATH!!!"""
		pass

	def on_after_set_atribute(self,name, value):
		"""Called after setting screenlet atributes"""
		pass

	def on_before_set_atribute(self,name, value):
		"""Called before setting screenlet atributes"""
		pass


	def on_create_drag_icon (self):
		"""Called when the screenlet's drag-icon is created. You can supply
		your own icon and mask by returning them as a 2-tuple."""
		return (None, None)

	def on_composite_changed(self):
		"""Called when composite state has changed"""
		pass

	def on_drag_begin (self, drag_context):
		"""Called when the Screenlet gets dragged."""
		pass
	
	def on_drag_enter (self, drag_context, x, y, timestamp):
		"""Called when something gets dragged into the Screenlets area."""
		pass
	
	def on_drag_leave (self, drag_context, timestamp):
		"""Called when something gets dragged out of the Screenlets area."""
		pass

	def on_drop (self, x, y, sel_data, timestamp):
		"""Called when a selection is dropped on this Screenlet."""
		return False
		
	def on_focus (self, event):
		"""Called when the Screenlet's window receives focus."""
		pass
	
	def on_hide (self):
		"""Called when the Screenlet gets hidden."""
		pass
	
	def on_init (self):
		"""Called when the Screenlet's options have been applied and the 
		screenlet finished its initialization. If you want to have your
		Screenlet do things on startup you should use this handler."""
		self.add_menuitem("update_now", "Refresh ...")
		# add default menu items
		self.add_default_menuitems()

		if self.verify_connection():
			self.calendar_login()

		self.update()
		
		# ADD a TIMER (in milliseconds) @ 15 mins
		self.timer = gobject.timeout_add( 1000 * 60 * 15, self.update)

	def on_key_down(self, keycode, keyvalue, event):
		"""Called when a keypress-event occured in Screenlet's window."""
		pass
	
	def on_load_theme (self):
		"""Called when the theme is reloaded (after loading, before redraw)."""
		pass
	
	def on_menuitem_select (self, id):
		"""Called when a menuitem is selected."""
		if id == "update_now":
			self.update()

		pass
	
	def on_mouse_down (self, event):
		"""Called when a buttonpress-event occured in Screenlet's window. 
		Returning True causes the event to be not further propagated."""
		return False
	
	def on_mouse_enter (self, event):
		"""Called when the mouse enters the Screenlet's window."""
		pass
		
	def on_mouse_leave (self, event):
		"""Called when the mouse leaves the Screenlet's window."""
		pass

	def on_mouse_move(self, event):
		"""Called when the mouse moves in the Screenlet's window."""
		self.redraw_canvas()
		pass

	def on_mouse_up (self, event):
		"""Called when a buttonrelease-event occured in Screenlet's window. 
		Returning True causes the event to be not further propagated."""
		return False
	
	def on_quit (self):
		"""Callback for handling destroy-event. Perform your cleanup here!"""
		print 'Bye'
		return True
		
	def on_realize (self):
		""""Callback for handling the realize-event."""
		pass
	
	def on_scale (self):
		"""Called when Screenlet.scale is changed."""
		pass
	
	def on_scroll_up (self):
		"""Called when mousewheel is scrolled up (button4)."""
		self.offset = max(0,self.offset-1)
		self.redraw_canvas()

	def on_scroll_down (self):
		"""Called when mousewheel is scrolled down (button5)."""
		self.offset = min(self.offset+1,len(self.appointment)-1)
		self.redraw_canvas()
	
	def on_show (self):
		"""Called when the Screenlet gets shown after being hidden."""
		pass
	
	def on_switch_widget_state (self, state):
		"""Called when the Screenlet enters/leaves "Widget"-state."""
		pass
	
	def on_unfocus (self, event):
		"""Called when the Screenlet's window loses focus."""
		pass
	
	def on_draw (self, ctx):
		# if theme is loaded
		if self.theme:
			ctx.scale(self.scale, self.scale)

			self.theme.render(ctx, self.backgr)

			ctx.save()
			ctx.translate(10,12)
			ctx.scale(.5,.5)
			self.theme.render(ctx, self.icon)
			ctx.restore()

			if self.appointment:	
				if len(self.appointment):
					ctx.translate(10,30)
					ctx.set_source_rgba(*self.font_color)
					self.draw_text(ctx, '<b><u>Agenda</u></b>', 0, 0, self.font , 12,self.width-35,pango.ALIGN_RIGHT)
					if not (self.on_line and self.logged_in):
						ctx.save()
						ctx.translate(0,16)
						ctx.set_source_rgba(*self.highlight_color_0)
						self.draw_text(ctx, '--offline--', 0, 0, self.font , 8,self.width-35,pango.ALIGN_RIGHT)
						ctx.restore()

					ctx.translate(7,40)
	
					step_size = 15
					line = 0
					# Iterating through the list of appointments.
					for a in range(self.offset,len(self.appointment)):
						if line > 13:
							break
						
						if a == self.offset or self.appointment[a][0] != self.appointment[a-1][0]:
							# draw boxes with alternating color.
							if line%2:
								ctx.set_source_rgba(*self.board_color_1)
							else:
								ctx.set_source_rgba(*self.board_color_0)
							self.draw_rectangle(ctx,0,-2,self.width-35,15)

							# print the date, if first entry to display or date
							# differs from date of previous appointment.
							ctx.set_source_rgba(*self.highlight_color_0)
							date = self.appointment[a][0].split('-')
							the_day = datetime.date(int(date[0]),int(date[1]),int(date[2]))
							self.draw_text(ctx, '<b>'+the_day.strftime('%a, %d. %b')+'</b>', 2, 0, self.font, 8, self.width-40, pango.ALIGN_LEFT)
							ctx.translate(0,step_size)
							line += 1

						if line%2:
							ctx.set_source_rgba(*self.board_color_1)
						else:
							ctx.set_source_rgba(*self.board_color_0)
						self.draw_rectangle(ctx,0,-2,self.width-35,15)
	
						# draw a square with calendar's color
						try:
							ctx.set_source_rgb(
								int(self.appointment[a][3][1:3],16)/255.,
								int(self.appointment[a][3][3:5],16)/255.,
								int(self.appointment[a][3][5:7],16)/255.
							)
						except:
							ctx.set_source_rgb(1.,1.,1.)
						self.draw_rectangle(ctx,4,2,7,7)

						# print the appointments time, if it is limited
						# to a given time.
						if self.appointment[a][1]:
							if len(self.appointment[a][1]) and self.appointment[a][1]!='None':
								ctx.set_source_rgba(*self.highlight_color_1)
								self.draw_text(ctx, self.appointment[a][1], 12, 0, self.font, 8, self.width-50, pango.ALIGN_LEFT)

						# print the title of the appointment.
						ctx.set_source_rgba(*self.font_color)
						if len(self.appointment[a][2]) > 20:
							self.draw_text(ctx, self.appointment[a][2][:17]+'...', 50, 0, self.font, 8, self.width-50, pango.ALIGN_LEFT)
						else:
							self.draw_text(ctx, self.appointment[a][2], 50, 0, self.font, 8, self.width-50, pango.ALIGN_LEFT)
						ctx.translate(0,step_size)
						line += 1
			else:
				ctx.translate(10,.4*self.height)
				ctx.set_source_rgba(*self.font_color)
				self.draw_text(ctx, 'No appointments\nto display.', 0, 0, self.font , 12,self.width-35,pango.ALIGN_CENTER)
		
	def on_draw_shape (self, ctx):
		self.on_draw(ctx)
	
# If the program is run directly or passed as an argument to the python
# interpreter then create a Screenlet instance and show it
if __name__ == "__main__":
	# create new session
	import screenlets.session
	screenlets.session.create_session(PhantCalendarScreenlet)

